/*
 * @file  Enumeration.h
 * @author  Nicolas Veyrat-Charvillon <nicolas.veyrat@uclouvain.be>
 * @date Mar 2, 2012
 * @version 1.0
 *
 * @section LICENSE
 *
 * Artistic License 2.0
 *
 * @ section DESCRIPTION
 *
 *  Class for enumerating key candidates
 *
 *	Coded using recursive templates to reflect the recursive
 *	nature of the efficient enumeration algorithm presented
 *	in [1].
 */

#ifndef ENUMERATION_H_
#define ENUMERATION_H_

#include "Event.h"
#include "EventQueue.h"
#include "EventList.h"
#include "Includes.h"

typedef array<vector<pair<unsigned char, double>>, 0x10> DistributionType;

template<unsigned N>
class Enumeration {
public:

	Enumeration(const array<vector<double> , N>& pDistribution);
	Enumeration(const array<vector<pair<unsigned char, double>>, N>& pDistribution);
	virtual ~Enumeration();

	pair<double, array<unsigned char, N>> nextCandidate();
	uint64_t memSize() const;
	uint64_t diskSize() const;
protected:
	Event<N> extractCandidate();
	bool addCandidate(uint64_t pX, uint64_t pY);

	Enumeration<N / 2> mEnumerationX;
	Enumeration<N / 2> mEnumerationY;

	EventQueue<N> mCandidates;

	EventList<N / 2> mListX;
	EventList<N / 2> mListY;
};

#include "EnumerationDefs.h"

template<>
class Enumeration<1> {
public:
Enumeration(const array<vector<pair<unsigned char, double>>, 1>& pDistribution) :
mCurrentIndex(0) {
	mDistribution = pDistribution[0];
	sort(mDistribution.begin(), mDistribution.end(),
			[](pair<unsigned char, double> v1, pair<unsigned char, double> v2) {return v1.second > v2.second;});

}

~Enumeration() {
}

pair<double, array<unsigned char, 1>> nextCandidate() {
	if (mCurrentIndex >= 0x100)
	throw runtime_error("Empty Candidate Set");
	double proba = mDistribution[mCurrentIndex].second;
	array<unsigned char, 1> index = { {mDistribution[mCurrentIndex].first}};
	++mCurrentIndex;
	return make_pair(proba, index);
}

uint64_t memSize() const {
	return sizeof(*this);
}

uint64_t diskSize() const {
	return 0;
}

protected:
vector<pair<unsigned char, double>> mDistribution;
uint64_t mCurrentIndex;
};

#endif /* ENUMERATION_H_ */
