/*
 * EventListHddDefs.h
 *
 *  Created on: Mar 29, 2012
 *      Author: nicolaq
 */

#ifndef EVENTLISTHDDDEFS_H_
#define EVENTLISTHDDDEFS_H_

template<unsigned N>
EventList<N>::EventList() :
	mEventTypeSize(sizeof(pair<double, array<unsigned char, N> > )), mSize(0),
			mBaseLimit(1 << 16), mCacheSize(0x100), mCacheMask(mCacheSize - 1),
			mLastEvents(mCacheSize) {
}

template<unsigned N>
EventList<N>::~EventList() {
	if (mEventStream.is_open()) {
		mEventStream.close();
		remove(fileName().c_str());
	}
}

template<unsigned N>
void EventList<N>::insert(const pair<double, array<unsigned char, N> >& pEvent) {
	mUsed.push_back(false);
	// stable memory storage
	if (mSize < mBaseLimit)
		mEventMemory.push_back(pEvent);
	else {
		mLastEvents[mSize & mCacheMask] = pEvent;
		// file storage
		openFile();
		if (mEventStream.is_open()) {
			mEventStream .seekp(mSize * mEventTypeSize);
			mEventStream.write(reinterpret_cast<const char*> (&pEvent),
					mEventTypeSize);
			extendCache();
		}
	}
	mSize++;
}

template<unsigned N>
uint64_t EventList<N>::size() const {
	return mSize;
}

template<unsigned N>
void EventList<N>::setMark(uint64_t pIndex) {
	mUsed[pIndex] = true;
}

template<unsigned N>
void EventList<N>::clearMark(uint64_t pIndex) {
	mUsed[pIndex] = false;
}

template<unsigned N>
bool EventList<N>::isMarked(uint64_t pIndex) const {
	return mUsed[pIndex];
}

template<unsigned N>
Event<N> EventList<N>::getEvent(uint64_t pIndex) const {
	// read Event from memory
	if (pIndex < mBaseLimit) {
		return Event<N> (pIndex, mEventMemory[pIndex]);
	}
	// read last inserted
	if (pIndex >= mSize - mCacheSize) {
		return Event<N> (pIndex, mLastEvents[pIndex & mCacheMask]);
	}
	uint64_t lowIndex = pIndex & mCacheMask;
	uint64_t highIndex = pIndex & ~mCacheMask;
	// read from cache
	auto& cachedValue = loadCache(highIndex);
	cachedValue.first = lowIndex;
	return Event<N> (pIndex, cachedValue.second[lowIndex]);
}

template<unsigned N>
string EventList<N>::fileName() const {
	stringstream s;
	s << "EventList_" << N << '@' << this;
	return s.str();
}

template<unsigned N>
uint64_t EventList<N>::memSize() const {
	uint64_t sizeValue = sizeof(*this);
	sizeValue += mEventMemory.size() * mEventTypeSize;
	sizeValue += (mUsed.size() + 7) >> 3;
	sizeValue += mCache.size() * (sizeof(pair<uint64_t, vector<pair<double,
			array<unsigned char, N> > > > ) + mCacheSize * mEventTypeSize);
	return sizeValue;
}

template<unsigned N>
uint64_t EventList<N>::diskSize() const {
	if (!mEventStream.is_open())
		return 0;
	return mSize * mEventTypeSize;
}

template<unsigned N>
void EventList<N>::openFile() const {
	if (mSize == mBaseLimit) {
		const auto flags = ios::in | ios::out | ios::binary | ios::trunc;
		mEventStream.open(fileName() + ".eventList", flags);
		mEventStream .write(
				reinterpret_cast<const char*> (&mEventMemory.front()),
				mEventMemory.size() * mEventTypeSize);
	}
}

template<unsigned N>
void EventList<N>::extendCache() {
	if (mCache.size() * mCacheSize > mBaseLimit) {
		mEventMemory.resize(mBaseLimit << 1);
		mEventStream .seekg(mBaseLimit * mEventTypeSize);
		mEventStream.read(reinterpret_cast<char*> (&mEventMemory[mBaseLimit]),
				mBaseLimit * mEventTypeSize);
		mBaseLimit <<= 1;
		auto it = mCache.begin();
		while (it != mCache.end())
			if (it->first < mBaseLimit)
				it = mCache.erase(it);
			else
				++it;
	}
}

template<unsigned N>
pair<uint64_t, vector<pair<double, array<unsigned char, N>>>>& EventList<N>::loadCache(uint64_t highIndex) const {
	auto& cachedValue = mCache[highIndex];
	// insert into cache if needed
	if (cachedValue.second.empty()) {
		auto previousCache = mCache.find(highIndex - mCacheSize);
		if (previousCache != mCache.end() && previousCache->second.first
				== mCacheMask) {
			cachedValue.second.swap(previousCache->second.second);
			mCache.erase(previousCache);
		}
		if (cachedValue.second.empty())
		cachedValue.second.resize(mCacheSize);
		mEventStream .seekg(highIndex * mEventTypeSize);
		mEventStream.read(
				reinterpret_cast<char*> (&cachedValue.second.front()),
				mCacheSize * mEventTypeSize);
	}
	return cachedValue;
}

#endif /* EVENTLISTHDDDEFS_H_ */
