/*
 * @file  keyEnumerationRank.cpp
 * @author  Nicolas Veyrat-Charvillon <nicolas.veyrat@uclouvain.be>
 * @date Jan 29, 2013
 * @version 1.1
 *
 * @section LICENSE
 *
 * Artistic License 2.0
 *
 * @ section DESCRIPTION
 *
 *  MATLAB hooks for key ranking using enumeration
 *
 * This is a MEX-file for MATLAB
 */

#include "mex.h"

#include "../Enumeration/Enumeration.h"

#include<algorithm>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	if (nrhs < 2 || nrhs > 3)
		mexErrMsgTxt(
				"Two arguments required: Event PMF, key value and optionally enumeration limit (2^24 by default)");
	if (nlhs != 1)
		mexErrMsgTxt("Too many output arguments");

	if (!(mxGetM(prhs[0]) == 0x100 && mxGetN(prhs[0]) == 0x10))
		mexErrMsgTxt("Event PMF must be 256 rows by 16 columns");
	if (!mxIsDouble(prhs[0]) || mxIsComplex(prhs[0]))
		mexErrMsgTxt("Event PMF must contain noncomplex scalar doubles");

	if (!(mxGetM(prhs[1]) == 0x1 && mxGetN(prhs[1]) == 0x10))
		mexErrMsgTxt("Key value must be 1 row by 16 columns");
	if (!mxIsUint8(prhs[1]))
		mexErrMsgTxt("Key value must contain 8-bit unsigned integers");

	unsigned long enumLimit = 1 << 24;
	if (nrhs == 3) {
		if (!(mxGetM(prhs[2]) == 1 && mxGetN(prhs[2]) == 1)
				|| !mxIsDouble(prhs[2]))
			mexErrMsgTxt("Third argument must be an unsigned integer");
		enumLimit = mxGetScalar(prhs[2]);
	}

	double *pPmfInput = mxGetPr(prhs[0]);

	array<vector<double>, 0x10> eventPMF;
	for(unsigned i0 = 0; i0 < 0x10; i0++)
		for(unsigned i1 = 0; i1 < 0x100; i1++)
			eventPMF[i0].push_back(pPmfInput[(i0 << 8) | i1]);

	Enumeration<0x10> enumeration(eventPMF);

	unsigned char* keyInput = static_cast<unsigned char*>(mxGetData(prhs[1]));
	array<unsigned char, 0x10> key;
	for(unsigned i0 = 0; i0 < 0x10; i0++)
		key[i0] = (unsigned char)keyInput[i0];

	mwSize outD = 2;
	mwSize outMN[2];
	outMN[0] = 1;
	outMN[1] = 1;
	plhs[0] = mxCreateNumericArray(outD, outMN, mxUINT64_CLASS, mxREAL);
	bool found = false;
	unsigned long long& rank = *static_cast<unsigned long long*>(mxGetData(plhs[0]));
	while (!found && rank < enumLimit) {
		auto candidate = enumeration.nextCandidate();
		found = std::equal(key.begin(), key.end(), candidate.second.begin());
		rank++;
	}
}
