/*
 * @file  rankEstimation.cpp
 * @author  Nicolas Veyrat-Charvillon <nicolas.veyrat@uclouvain.be>
 * @date Jan 29, 2013
 * @version 1.0
 *
 * @section LICENSE
 *
 * Artistic License 2.0
 *
 * @ section DESCRIPTION
 *
 *  MATLAB hooks for key rank estimation
 *
 * This is a MEX-file for MATLAB
 */

#include "mex.h"

#include "../Rank/RankKey.h"

#include<algorithm>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	if (nrhs < 3 || nrhs > 4)
		mexErrMsgTxt(
				"Two arguments required: Event PMF, key probability, enumeration time and optionally log2 of merging limit (24 by default)");
	if (nlhs != 1)
		mexErrMsgTxt("Too many output arguments");

	if (!(mxGetM(prhs[0]) == 0x100 && mxGetN(prhs[0]) == 0x10))
		mexErrMsgTxt("Event PMF must be 256 rows by 16 columns");
	if (!mxIsDouble(prhs[0]) || mxIsComplex(prhs[0]))
		mexErrMsgTxt("Event PMF must contain noncomplex scalar doubles");

		if (!(mxGetM(prhs[1]) == 1 && mxGetN(prhs[1]) == 1)
				&& mxIsDouble(prhs[1]))
		mexErrMsgTxt("Key probability must be a real value");

	double keyProbability = static_cast<double>(mxGetScalar(prhs[1]));

		if (!(mxGetM(prhs[2]) == 1 && mxGetN(prhs[2]) == 1)
				&& !mxIsDouble(prhs[2]))
			mexErrMsgTxt("Estimation time must be an unsigned");
	double estimationTime = static_cast<unsigned>(mxGetScalar(prhs[2]));

	unsigned long merging = 24;
	if (nrhs == 4) {
		if (!(mxGetM(prhs[3]) == 1 && mxGetN(prhs[3]) == 1)
				|| !mxIsDouble(prhs[3]))
			mexErrMsgTxt("Merging limit must be an unsigned integer");
		merging = mxGetScalar(prhs[3]);
	}

	double *pPmfInput = mxGetPr(prhs[0]);

	vector<vector<double>> eventPMF(0x10);
	for(unsigned i0 = 0; i0 < 0x10; i0++)
		for(unsigned i1 = 0; i1 < 0x100; i1++)
			eventPMF[i0].push_back(pPmfInput[(i0 << 8) | i1]);

	RankKey rankEstim(eventPMF, keyProbability);
	rankEstim.mergeDistribution(merging);
	auto estimation = rankEstim(estimationTime);

	mwSize outD = 2;
	mwSize outMN[2];
	outMN[0] = 1;
	outMN[1] = 3;
	plhs[0] = mxCreateNumericArray(outD, outMN, mxDOUBLE_CLASS, mxREAL);
	double *rank = static_cast<double *>(mxGetData(plhs[0]));
	for(unsigned i0 = 0; i0 < 3; ++i0)
		rank[i0] = log(double(estimation[i0])) / log(2);
}
