/*
 * RankKey.cpp
 *
 *  Created on: May 14, 2012
 *      Author: nicolaq
 */

#include "RankKey.h"

RankKey::RankKey(const vector<vector<double>>& pDistribution,
		double pKeyProbability) :
		mDistribution(pDistribution), mKeyProbability(pKeyProbability) {
	mEstimation.fill(0);
	mEstimation[0] = 1;
	for (auto& v0 : mDistribution)
		mEstimation[0] *= v0.size();
}

RankKey::~RankKey() {
}

void RankKey::mergeDistribution(unsigned logSize) {
	uint128 maxSize = 1;
	maxSize <<= logSize;
	vector<vector<double>> mergedPmf;
	for (auto& v0 : mDistribution) {
		if ((mergedPmf.size() == 0)
				|| (mergedPmf.back().size() * v0.size() > maxSize))
			mergedPmf.push_back(vector<double>(1, 1));
		vector<double> merged;
		for (auto& v1 : v0)
			for (auto& v2 : mergedPmf.back())
				merged.push_back(v1 * v2);
		mergedPmf.back() = merged;
	}
	mDistribution = mergedPmf;
}

array<uint128, 3> RankKey::operator()(unsigned pSeconds) {
	sortDistribution();
	mTopes.push(KeyTope(mDistribution));
	auto predicate = [this](const vector<unsigned>& point) {
		double diff = this->probability(point) - this->mKeyProbability;
		return diff > 0 ? 1 : (diff < 0 ? -1 : 0);};
	auto startTime = chrono::system_clock::now();
	while (!mTopes.empty() && chrono::duration<double>(chrono::system_clock::now() - startTime).count()
			< pSeconds) {
		KeyTope tope = mTopes.pop();
		if (tope.isCarved()) {
			KeyTope splitTope = tope.splitLargest();
			mTopes.push(tope);
			mTopes.push(splitTope);
		} else {
			auto carvedVolume = tope.carvePiece(predicate, mEstimation);
			if (tope.carvedSign() < 0)
				mEstimation[tope.carvedSign() + 1] -= carvedVolume;
			else
				mEstimation[tope.carvedSign() + 1] += carvedVolume;
			if (tope.volume() > 0) {
				mTopes.push(tope);
			}
		}
	}
	return mEstimation;
}

void RankKey::sortDistribution() {
	for (auto& v0 : mDistribution)
		sort(v0.begin(), v0.end(), [](double v1, double v2) {return v1 > v2;});
}

double RankKey::probability(const vector<unsigned>& pPoint) const {
	double probability = 1;
	for (unsigned i0 = 0; i0 < pPoint.size(); ++i0)
		probability *= mDistribution[i0][pPoint[i0]];
	return probability;
}
